import * as log4js from 'log4js';
import * as promises from 'node:fs/promises';
import * as path from 'node:path';
import { XMLBuilder, XMLParser } from 'fast-xml-parser';
import * as fs from 'node:fs';
import { createHash } from 'node:crypto';

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise, SuppressedError, Symbol */


function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
    var e = new Error(message);
    return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
};

const XML_HEADER = '<?xml version="1.0" encoding="UTF-8"?>';
const INDENT = "    ";

const XML_PARSER_OPTION = {
    commentPropName: "!---",
    ignoreAttributes: false,
    ignoreNameSpace: false,
    parseTagValue: false,
    parseNodeValue: false,
    parseAttributeValue: false,
    trimValues: true,
    processEntities: false,
    cdataPropName: "![CDATA[",
};
const JSON_PARSER_OPTION = Object.assign(Object.assign({}, XML_PARSER_OPTION), { format: true, indentBy: INDENT, suppressBooleanAttributes: false, suppressEmptyNode: false });

function buildReassembledFile(combinedXmlContents, filePath, xmlElement, xmlRootElementHeader) {
    return __awaiter(this, void 0, void 0, function* () {
        let finalXmlContent = combinedXmlContents.join("\n");
        finalXmlContent = finalXmlContent.replace(/<\?xml version="1.0" encoding="UTF-8"\?>/g, "");
        finalXmlContent = finalXmlContent.replace(new RegExp(`<${xmlElement}\\s*[^>]*>`, "g"), "");
        finalXmlContent = finalXmlContent.replace(new RegExp(`</${xmlElement}>`, "g"), "");
        finalXmlContent = finalXmlContent.replace(/<!\[CDATA\[\s*([\s\S]*?)\s*]]>/g, function (_, cdataContent) {
            const trimmedContent = cdataContent.trim();
            const lines = trimmedContent.split("\n");
            const indentedLines = lines.map(function (line) {
                return line.replace(/^\s*/, "");
            });
            return ("<![CDATA[\n" + INDENT + indentedLines.join("\n" + INDENT) + "\n]]>");
        });
        finalXmlContent = finalXmlContent.replace(/(\n\s*){2,}/g, `\n${INDENT}`);
        const closeTag = `</${xmlElement}>`;
        yield promises.writeFile(filePath, `${XML_HEADER}\n${xmlRootElementHeader}${finalXmlContent}${closeTag}`);
        logger.debug(`Created reassembled file: ${filePath}`);
    });
}

function buildRootElementHeader(rootElement, rootElementName) {
    let rootElementHeader = `<${rootElementName}`;
    for (const [attrKey, attrValue] of Object.entries(rootElement)) {
        if (attrKey.startsWith("@")) {
            const cleanAttrKey = attrKey.slice(2);
            rootElementHeader += ` ${cleanAttrKey}="${String(attrValue)}"`;
        }
    }
    rootElementHeader += ">";
    return rootElementHeader;
}

function buildXMLString(element, indentLevel = 0) {
    const xmlBuilder = new XMLBuilder(JSON_PARSER_OPTION);
    const xmlString = xmlBuilder.build(element);
    const formattedXml = xmlString
        .split("\n")
        .map((line) => `${" ".repeat(indentLevel * INDENT.length)}${line}`)
        .join("\n")
        .trimEnd();
    return formattedXml;
}

const xmlParser = new XMLParser(XML_PARSER_OPTION);
class ReassembleXMLFileHandler {
    processFilesInDirectory(dirPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const combinedXmlContents = [];
            let rootResult = undefined;
            const files = yield promises.readdir(dirPath);
            files.sort((fileA, fileB) => {
                const fullNameA = fileA.split(".")[0].toLowerCase();
                const fullNameB = fileB.split(".")[0].toLowerCase();
                return fullNameA.localeCompare(fullNameB);
            });
            for (const file of files) {
                const filePath = path.join(dirPath, file);
                const fileStat = yield promises.stat(filePath);
                if (fileStat.isFile() && filePath.endsWith(".xml")) {
                    const xmlContent = yield promises.readFile(filePath, "utf-8");
                    let xmlParsed;
                    try {
                        xmlParsed = xmlParser.parse(xmlContent, true);
                    }
                    catch (err) {
                        logger.error(`${filePath} was unable to be parsed and was not added to the reassembled file. Confirm formatting and try again.`);
                        continue;
                    }
                    const rootResultFromFile = yield this.processFilesForRootElement(xmlParsed);
                    if (rootResultFromFile && !rootResult) {
                        rootResult = rootResultFromFile;
                    }
                    const combinedXmlString = buildXMLString(xmlParsed);
                    combinedXmlContents.push(combinedXmlString);
                }
                else if (fileStat.isDirectory()) {
                    const [subCombinedXmlContents, subRootResult] = yield this.processFilesInDirectory(filePath);
                    combinedXmlContents.push(...subCombinedXmlContents);
                    if (subRootResult && !rootResult) {
                        rootResult = subRootResult;
                    }
                }
            }
            return [combinedXmlContents, rootResult];
        });
    }
    processFilesForRootElement(xmlParsed) {
        return __awaiter(this, void 0, void 0, function* () {
            const rootElementName = Object.keys(xmlParsed)[1];
            const rootElement = xmlParsed[rootElementName];
            const rootElementHeader = buildRootElementHeader(rootElement, rootElementName);
            return [rootElementName, rootElementHeader];
        });
    }
    reassemble(xmlAttributes) {
        return __awaiter(this, void 0, void 0, function* () {
            const { xmlPath, fileExtension, postPurge = false } = xmlAttributes;
            let combinedXmlContents = [];
            const fileStat = yield promises.stat(xmlPath);
            if (!fileStat.isDirectory()) {
                logger.error(`The provided xmlPath ${xmlPath} to reassemble is not a directory.`);
                return;
            }
            logger.debug(`Parsing directory to reassemble: ${xmlPath}`);
            const [subCombinedXmlContents, rootResult] = yield this.processFilesInDirectory(xmlPath);
            combinedXmlContents = subCombinedXmlContents;
            const parentDirectory = path.dirname(xmlPath);
            const subdirectoryBasename = path.basename(xmlPath);
            const fileName = fileExtension
                ? `${subdirectoryBasename}.${fileExtension}`
                : `${subdirectoryBasename}.xml`;
            const filePath = path.join(parentDirectory, fileName);
            if (rootResult !== undefined) {
                const [rootElementName, rootElementHeader] = rootResult;
                yield buildReassembledFile(combinedXmlContents, filePath, rootElementName, rootElementHeader);
                if (postPurge)
                    yield promises.rm(xmlPath, { recursive: true });
            }
            else {
                logger.error(`No files under ${xmlPath} were parsed successfully. A reassembled XML file was not created.`);
            }
        });
    }
}

function findUniqueIdElement(element, uniqueIdElements) {
    if (uniqueIdElements === undefined) {
        return getShortHash(element);
    }
    const uniqueIdElementsArray = uniqueIdElements.split(",");
    for (const fieldName of uniqueIdElementsArray) {
        if (element[fieldName] !== undefined) {
            if (typeof element[fieldName] === "string") {
                return element[fieldName];
            }
        }
    }
    for (const key in element) {
        if (typeof element[key] === "object" && element[key] !== null) {
            const childFieldName = findUniqueIdElement(element[key], uniqueIdElements);
            if (childFieldName !== undefined) {
                return childFieldName;
            }
        }
    }
    return getShortHash(element);
}
function getShortHash(element) {
    const hash = createHash("sha256");
    hash.update(JSON.stringify(element));
    const fullHash = hash.digest("hex");
    return fullHash.slice(0, 8);
}

function buildNestedFile(element, metadataPath, uniqueIdElements, rootElementName, rootElementHeader, parentKey, indent) {
    return __awaiter(this, void 0, void 0, function* () {
        let elementContent = "";
        const fieldName = findUniqueIdElement(element, uniqueIdElements);
        const outputDirectory = path.join(metadataPath, parentKey);
        const outputFileName = `${fieldName}.${parentKey}-meta.xml`;
        const outputPath = path.join(outputDirectory, outputFileName);
        yield promises.mkdir(outputDirectory, { recursive: true });
        const parentKeyHeader = buildRootElementHeader(element, parentKey);
        elementContent = buildXMLString(element, 2);
        let decomposeFileContents = `${XML_HEADER}\n`;
        decomposeFileContents += `${rootElementHeader}\n`;
        decomposeFileContents += `${indent}${parentKeyHeader}\n`;
        decomposeFileContents += `${elementContent}\n`;
        decomposeFileContents += `${indent}</${parentKey}>\n`;
        decomposeFileContents += `</${rootElementName}>`;
        yield promises.writeFile(outputPath, decomposeFileContents);
        logger.debug(`Created disassembled file: ${outputPath}`);
    });
}

function processElement(params) {
    return __awaiter(this, void 0, void 0, function* () {
        const { element, metadataPath, uniqueIdElements, rootElementName, rootElementHeader, key, indent, leafContent, leafCount, hasNestedElements, } = params;
        if (typeof element === "object") {
            yield buildNestedFile(element, metadataPath, uniqueIdElements, rootElementName, rootElementHeader, key, indent);
            return [leafContent, leafCount, true];
        }
        else {
            const fieldValue = element;
            const updatedLeafContent = `${leafContent}${indent}<${key}>${String(fieldValue)}</${key}>\n`;
            return [updatedLeafContent, leafCount + 1, hasNestedElements];
        }
    });
}

function buildLeafFile(leafContent, metadataPath, baseName, rootElementName, rootElementHeader) {
    return __awaiter(this, void 0, void 0, function* () {
        let leafFile = `${XML_HEADER}\n`;
        leafFile += `${rootElementHeader}\n`;
        const sortedLeafContent = leafContent
            .split("\n")
            .filter((line) => line.trim() !== "")
            .sort((a, b) => a.localeCompare(b))
            .join("\n");
        leafFile += sortedLeafContent;
        leafFile += `\n</${rootElementName}>`;
        const leafOutputPath = path.join(metadataPath, `${baseName}.xml`);
        yield promises.writeFile(leafOutputPath, leafFile);
        logger.debug(`Created disassembled file: ${leafOutputPath}`);
    });
}

function buildDisassembledFiles(xmlString, metadataPath, uniqueIdElements, baseName, indent, postPurge, parentPath) {
    return __awaiter(this, void 0, void 0, function* () {
        const xmlParser = new XMLParser(XML_PARSER_OPTION);
        let result;
        try {
            result = xmlParser.parse(xmlString, true);
        }
        catch (err) {
            logger.error(`${baseName}.xml was unable to be parsed. Confirm formatting and try again.`);
            return;
        }
        const rootElementName = Object.keys(result)[1];
        const rootElement = result[rootElementName];
        const rootElementHeader = buildRootElementHeader(rootElement, rootElementName);
        let leafContent = "";
        let leafCount = 0;
        let hasNestedElements = false;
        for (const key of Object.keys(rootElement).filter((key) => !key.startsWith("@"))) {
            if (Array.isArray(rootElement[key])) {
                for (const element of rootElement[key]) {
                    const [updatedLeafContent, updatedLeafCount, updatedHasNestedElements] = yield processElement({
                        element,
                        metadataPath,
                        uniqueIdElements,
                        rootElementName,
                        rootElementHeader,
                        key,
                        indent,
                        leafContent,
                        leafCount,
                        hasNestedElements,
                    });
                    leafContent = updatedLeafContent;
                    leafCount = updatedLeafCount;
                    hasNestedElements = updatedHasNestedElements;
                }
            }
            else {
                const [updatedLeafContent, updatedLeafCount, updatedHasNestedElements] = yield processElement({
                    element: rootElement[key],
                    metadataPath,
                    uniqueIdElements,
                    rootElementName,
                    rootElementHeader,
                    key,
                    indent,
                    leafContent,
                    leafCount,
                    hasNestedElements,
                });
                leafContent = updatedLeafContent;
                leafCount = updatedLeafCount;
                hasNestedElements = updatedHasNestedElements;
            }
        }
        if (!hasNestedElements) {
            logger.error(`The XML file ${baseName}.xml only has leaf elements. This file will not be disassembled.`);
            return;
        }
        if (leafCount > 0) {
            yield buildLeafFile(leafContent, metadataPath, baseName, rootElementName, rootElementHeader);
        }
        if (postPurge) {
            const originalFilePath = path.resolve(`${parentPath}/${baseName}.xml`);
            promises.unlink(originalFilePath);
        }
    });
}

class DisassembleXMLFileHandler {
    disassemble(xmlAttributes) {
        return __awaiter(this, void 0, void 0, function* () {
            const { xmlPath, uniqueIdElements, prePurge = false, postPurge = false, } = xmlAttributes;
            const fileStat = yield promises.stat(xmlPath);
            if (fileStat.isFile()) {
                const filePath = path.resolve(xmlPath);
                if (!filePath.endsWith(".xml")) {
                    logger.error(`The file path ${filePath} is not an XML file.`);
                    return;
                }
                const basePath = path.dirname(filePath);
                yield this.processFile({
                    xmlPath: basePath,
                    filePath,
                    uniqueIdElements,
                    prePurge,
                    postPurge,
                });
            }
            else if (fileStat.isDirectory()) {
                const files = yield promises.readdir(xmlPath);
                for (const file of files) {
                    const filePath = path.join(xmlPath, file);
                    if (filePath.endsWith(".xml")) {
                        yield this.processFile({
                            xmlPath,
                            filePath,
                            uniqueIdElements,
                            prePurge,
                            postPurge,
                        });
                    }
                }
            }
        });
    }
    processFile(xmlAttributes) {
        return __awaiter(this, void 0, void 0, function* () {
            const { xmlPath, filePath, uniqueIdElements, prePurge, postPurge } = xmlAttributes;
            logger.debug(`Parsing file to disassemble: ${filePath}`);
            const xmlContent = yield promises.readFile(filePath, "utf-8");
            const fullName = path.basename(filePath, path.extname(filePath));
            const baseName = fullName.split(".")[0];
            let outputPath;
            outputPath = path.join(xmlPath, baseName);
            if (prePurge) {
                if (fs.existsSync(outputPath)) {
                    yield promises.rm(outputPath, { recursive: true });
                }
            }
            yield buildDisassembledFiles(xmlContent, outputPath, uniqueIdElements, fullName, INDENT, postPurge, xmlPath);
        });
    }
}

function setLogLevel(level) {
    log4js.getLogger().level = level;
}
const logger = log4js.getLogger();
log4js.configure({
    appenders: { disassemble: { type: "file", filename: "disassemble.log" } },
    categories: { default: { appenders: ["disassemble"], level: "error" } },
});

export { DisassembleXMLFileHandler, ReassembleXMLFileHandler, logger, setLogLevel };
//# sourceMappingURL=index.mjs.map
